import argparse
import os
import pandas as pd
import warnings

import libsbml as sbml
import petab
import yaml


from .yaml2sbml import _parse_yaml_dict, _load_yaml_file
from .yaml_validation import _validate_yaml_from_dict


def yaml2petab(yaml_file: str,
               output_dir: str,
               model_name: str,
               petab_yaml_name: str = None,
               measurement_table_name: str = None):
    """
    Takes in a yaml file with the ODE specification, parses it, converts
    it into SBML format, and writes the SBML file. Further it translates the
    given information into PEtab tables.

    If a petab_yaml_name is given, a .yaml file is created, that organizes
    the petab problem. If additionally a measurement_table_file_name is
    specified, this file name is written into the created .yaml file.

    Arguments:
        yaml_file : path to the yaml file with the ODEs specification
        output_dir: path the output file(s) are be written out
        model_name: name of SBML model
        petab_yaml_name: name of yaml organizing the PEtab problem.
        measurement_table_name: Name of measurement table

    Returns:

    Raises:

    """
    yaml_model_dict = _load_yaml_file(yaml_file)
    _yaml2petab(yaml_model_dict,
                output_dir,
                model_name,
                petab_yaml_name,
                measurement_table_name)


def _yaml2petab(yaml_model_dict: dict,
                output_dir: str,
                model_name: str,
                petab_yaml_name: str = None,
                measurement_table_name: str = None):
    """
    Similar to 'yaml2petab', but takes a yaml_model_dict as input,
    instead of a file path.

    Takes in a yaml model dict with the ODE specification, parses it, converts
    it into SBML format, and writes the SBML file. Further it translates the
    given information into PEtab tables.

    If a petab_yaml_name is given, a .yaml file is created, that organizes
    the petab problem. If additionally a measurement_table_file_name is
    specified, this file name is written into the created .yaml file.

    Arguments:
        yaml_model_dict : dictionary, containing the yaml model
        output_dir: path the output file(s) are be written out
        model_name: name of SBML model
        petab_yaml_name: name of yaml organizing the PEtab problem.
        measurement_table_name: Name of measurement table

    Returns:

    Raises:

    """
    # validate yaml
    _validate_yaml_from_dict(yaml_model_dict)

    # output make directory, if it doesn't exist yet.
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)

    if model_name.endswith('.xml') or model_name.endswith('.sbml'):
        sbml_dir = os.path.join(output_dir, model_name)
    else:
        sbml_dir = os.path.join(output_dir, model_name + '.xml')

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        sbml_as_string = _parse_yaml_dict(yaml_model_dict)

    with open(sbml_dir, 'w') as f_out:
        f_out.write(sbml_as_string)

    # create petab tsv files:
    _create_petab_tables_from_yaml(yaml_model_dict, output_dir)

    # create yaml file, that organizes the petab problem:
    if (petab_yaml_name is None) and (measurement_table_name is not None):

        warnings.warn('Since no petab_yaml_file_name is specified, the '
                      'specified measurement_table_name will have no effect.',
                      RuntimeWarning)

    elif petab_yaml_name is not None:
        _create_petab_problem_yaml(yaml_model_dict,
                                   output_dir,
                                   sbml_dir,
                                   petab_yaml_name,
                                   measurement_table_name)
    # validate PEtab tables:
    validate_petab_tables(sbml_dir, output_dir)


def _create_petab_tables_from_yaml(yaml_dict: dict,
                                   output_dir: str):
    """
    Parses the yaml dict to a PEtab observable/parameter table.

    Arguments:
        yaml_dict: dict, that contains the yaml file.
        output_dir: directory, where the PEtab tables should be written.

    Raises:

    """
    parameter_table = _create_parameter_table(yaml_dict)
    parameter_table.to_csv(os.path.join(output_dir, 'parameter_table.tsv'),
                           sep='\t',
                           index=False)

    # create PEtab observable table, if observables occur in the yaml file.
    if 'observables' in yaml_dict.keys():
        observable_table = _create_observable_table(yaml_dict)
        observable_table.to_csv(os.path.join(output_dir,
                                             'observable_table.tsv'),
                                sep='\t',
                                index=False)

    # create PEtab condition table, if conditions occur in the yaml file.
    if 'conditions' in yaml_dict.keys():
        condition_table = _create_condition_table(yaml_dict)
        condition_table.to_csv(os.path.join(output_dir, 'condition_table.tsv'),
                               sep='\t',
                               index=False)


def _create_petab_problem_yaml(yaml_dict: dict,
                               output_dir: str,
                               sbml_dir: str,
                               petab_yaml_name: str,
                               measurement_table_name: str = None):
    """
    Creates the yaml file, that can be used for structuring a PEtab problem in
    the corresponding directory.

    Arguments:
        yaml_dict: dict, that contains the yaml file.
        output_dir: directory, where the PEtab tables should be written.
        sbml_dir: directory of the SBML model.
        petab_yaml_name: name of file, where PEtab yaml is written.
        measurement_table_name: directory of the  measurement table.

    Raises:

    """
    petab_yaml_dict = {'format_version': 1,
                       'parameter_file': 'parameter_table.tsv',
                       'problems': [{'sbml_files':
                                         [os.path.basename(sbml_dir)]}]}

    # fill the corresponding entries, if they are contained in the yaml/input.

    if 'observables' in yaml_dict.keys():
        petab_yaml_dict['problems'][0]['observable_files'] = \
            ['observable_table.tsv']

    if 'conditions' in yaml_dict.keys():
        petab_yaml_dict['problems'][0]['condition_files'] = \
            ['condition_table.tsv']

    if measurement_table_name is not None:
        petab_yaml_dict['problems'][0]['measurement_files'] = \
            [measurement_table_name]

    petab_yaml_dir = os.path.join(output_dir, petab_yaml_name)

    with open(petab_yaml_dir, 'w') as file:
        yaml.dump(petab_yaml_dict, file)


def _create_parameter_table(yaml_dict: dict):
    """
    Creates a parameter table from the parameter block in the given yaml_dict.

    Arguments:
        yaml_dict

    Returns:
        parameter_table: pandas data frame containing the parameter table.

    Raises:

    """

    return _create_petab_table(yaml_dict['parameters'],
                               petab.PARAMETER_DF_REQUIRED_COLS,
                               petab.PARAMETER_DF_OPTIONAL_COLS)


def _create_observable_table(yaml_dict: dict):
    """
    Creates an observable table from the observable block
    in the given yaml_dict.

    Arguments:
        yaml_dict

    Returns:
        observable_table: pandas data frame containing the observable table.
            (if observable block is not empty, else None)

    Raises:

    """

    return _create_petab_table(yaml_dict['observables'],
                               petab.OBSERVABLE_DF_REQUIRED_COLS,
                               petab.OBSERVABLE_DF_OPTIONAL_COLS)


def _create_condition_table(yaml_dict: dict):
    """
    Creates a condition table from the observable block in the given yaml_dict.

    Arguments:
        yaml_dict

    Returns:
        condition_table: pandas data frame containing the condition table.
            (if condition block is not empty, else None)

    Raises:

    """
    mandatory_id_list = [petab.CONDITION_ID]

    optional_id_list = [petab.CONDITION_NAME] + \
                       [p['parameterId'] for p in yaml_dict['parameters']] + \
                       [ode['stateId'] for ode in yaml_dict['odes']]

    return _create_petab_table(yaml_dict['conditions'],
                               mandatory_id_list,
                               optional_id_list)


def validate_petab_tables(sbml_dir: str, output_dir: str):
    """
    Validates the petab tables via petab.lint. Throws an error,
    if the petab tables do not follow petab format standard.

    Arguments:
        sbml_dir: directory of the sbml
        output_dir: output directory for petab files

    Returns:

    Raises:
        Errors are raised by lint, if PEtab files are invalid...

    """
    model = sbml.readSBML(sbml_dir).getModel()

    parameter_file_dir = os.path.join(output_dir, 'parameter_table.tsv')
    observable_file_dir = os.path.join(output_dir, 'observable_table.tsv')
    condition_table_dir = os.path.join(output_dir, 'condition_table.tsv')

    # check observable table, if the table exists
    if os.path.exists(observable_file_dir):
        observable_df = pd.read_csv(observable_file_dir,
                                    sep='\t',
                                    index_col='observableId')
        petab.lint.check_observable_df(observable_df)

    else:
        observable_df = None

    # check condition table, if the table exists
    if os.path.exists(condition_table_dir):
        condition_df = pd.read_csv(condition_table_dir,
                                   sep='\t',
                                   index_col='conditionId')
        petab.lint.check_condition_df(condition_df, model)

    # check parameter table
    parameter_df = pd.read_csv(parameter_file_dir,
                               sep='\t',
                               index_col='parameterId')
    petab.lint.check_parameter_df(parameter_df,
                                  sbml_model=model,
                                  observable_df=observable_df)


def _create_petab_table(block_list: list,
                        mandatory_id_list: list,
                        optional_id_list: id):
    """
    Creates a PEtab table from the block_list in the yaml_dict.

    Arguments:
        block_list: entry from yaml_dict.
        mandatory_id_list: list of mandatory ids in the PEtab table
        optional_id_list: list of optional ids in the PEtab table

    Returns:
        petab_table: pandas data frame containing the petab table.

    Raises:

    """

    petab_table = pd.DataFrame({col_id: [] for col_id in mandatory_id_list})

    for row in block_list:
        _petab_table_add_row(petab_table, row)

    # check if every column is part of PEtab standard.
    for col_name in petab_table.head():
        if not (col_name in mandatory_id_list or col_name in optional_id_list):
            warnings.warn(f'PEtab warning: {col_name} is not part of the '
                          f'PEtab standard and hence might have noe effect.')
    return petab_table


def _petab_table_add_row(petab_table: pd.DataFrame, row_dict: dict):
    """
    Adds a row defined by row_dict to the petab_table given by petab_table.
    In the end the last row of petab_table has the entries

        petab_table[-1, key] = row_dict[key]

    for all keys of row_dict.

    Arguments:
        petab_table: pd.DataFrame containing the current petab table.
        row_dict: dict, that contains the values of the new row.

    Returns:

    Raises:

    """
    n_rows = petab_table.shape[0]

    for key in row_dict.keys():
        petab_table.loc[n_rows + 1, key] = row_dict[key]


def main():

    parser = argparse.ArgumentParser(description='Takes in an ODE model in '
                                     '.yaml and converts it to a PEtab file.')

    parser.add_argument('yaml_file', type=str, help='Input yaml directory')
    parser.add_argument('output_dir', type=str, help='Output directory')
    parser.add_argument('model_name', type=str,
                        help='name of created SBML model.')
    parser.add_argument('-y', '--petab_yaml', type=str,
                        help='Optional argument, creates a petab .yml')
    parser.add_argument('-m', '--measurement_table', type=str,
                        help='Optional argument, path to measurement table')

    args = parser.parse_args()

    print(f'Path to yaml file: {args.yaml_file}')
    print(f'Output directory: {args.output_dir}')
    print(f'Path to sbml/petab files: {args.model_name}')

    print('Converting...')

    yaml2petab(args.yaml_file,
               args.output_dir,
               args.model_name,
               args.petab_yaml,
               args.measurement_table)


if __name__ == '__main__':
    main()



import argparse
import warnings

import libsbml as sbml
import yaml

from .yaml_validation import _validate_yaml_from_dict


def yaml2sbml(yaml_file: str, sbml_file: str):
    """
    Takes in a yaml file with the ODE specification, parses it, converts it
    into SBML format, and writes the SBML file.

    Arguments:
        yaml_file : path to the yaml file with the ODEs specification
        sbml_file: path to the SBML file to be written out

    Returns:

    Raises:

    """
    sbml_as_string = _parse_yaml(yaml_file)

    # write sbml file
    with open(sbml_file, 'w') as f_out:
        f_out.write(sbml_as_string)


def _parse_yaml(yaml_file: str) -> str:
    """
    Takes in a yaml file with the specification of ODEs, parses it, and
    returns the corresponding SBML string.

    Arguments:
        yaml_file: path to the yaml file with the ODEs specification

    Returns:
        sbml_string: a string containing the ODEs in SBML format

    Raises:
        SystemExit
    """
    yaml_dict = _load_yaml_file(yaml_file)
    _validate_yaml_from_dict(yaml_dict)
    sbml_string = _parse_yaml_dict(yaml_dict)

    return sbml_string


def _parse_yaml_dict(yaml_dict) -> str:
    """
    Generates a string, containing the SBML from a yaml_dict.

    Arguments:
        yaml_dict: dictionary, containing to the yaml file with the ODEs
                   specification.

    Returns:
        sbml_string: a string containing the ODEs in SBML format.

    """

    try:
        document = sbml.SBMLDocument(3, 1)
    except ValueError:
        raise SystemExit('Could not create SBMLDocument object')

    model = document.createModel()
    model = _create_compartment(model)

    _convert_yaml_blocks_to_sbml(model, yaml_dict)

    # check consistency and give warnings for errors in SBML:
    if document.checkConsistency():

        for error_num in range(document.getErrorLog().getNumErrors()):
            if not document.getErrorLog().getError(error_num).isWarning():
                warnings.warn(
                    document.getErrorLog().getError(error_num).getMessage(),
                    RuntimeWarning)

    sbml_string = sbml.writeSBMLToString(document)

    return sbml_string


def _create_compartment(model: sbml.Model):
    """
    Creates a default compartment for the model.
    We don't support multiple compartments at the moment.

    Arguments:
        model: SBML model

    Returns:
        model: SBML model with added compartment

    Raises:

    """
    c = model.createCompartment()
    c.setId('Compartment')
    c.setConstant(True)
    c.setSize(1)

    return model


def _load_yaml_file(yaml_file: str) -> dict:
    """
    Loads yaml file and returns the resulting dictionary.

    Arguments:
        yaml_file: SBML model

    Returns:
        yaml_dic: dictionary with parsed yaml file contents

    Raises:

    """
    with open(yaml_file, 'r') as f_in:
        yaml_contents = f_in.read()
        yaml_dic = yaml.full_load(yaml_contents)

    return yaml_dic


def _convert_yaml_blocks_to_sbml(model: sbml.Model, yaml_dic: dict):
    """
    Converts each block in the yaml dictionary to SBML.

    Arguments:
        model: SBML model
        yaml_dic: dictionary with yaml contents

    Returns:
        model: SBML model with added entities

    Raises:

    """
    function_dict = {'time': _read_time_block,
                     'parameters': _read_parameters_block,
                     'assignments': _read_assignments_block,
                     'functions': _read_functions_block,
                     'observables': _read_observables_block,
                     'odes': _read_odes_block,
                     'conditions': _read_conditions_block}

    for block in yaml_dic:
        function_dict[block](model, yaml_dic[block])

    return model


def _read_time_block(model: sbml.Model, time_dic: dict):
    """
    Reads and processes the time block.

    Arguments:
        model: SBML model to which the rate rule will be added.
        time_dic: a dictionary with the time block in the ODE yaml file.

    Returns:

    Raises:

    """

    if time_dic['variable'] == 'time':
        return
    else:
        _create_time(model, time_dic['variable'])


def _create_time(model: sbml.Model, time_var: str):
    """
    Creates the time variable, add assignment to 'time'

    Arguments:
        model: the SBML model to which the species will be added.
        time_var: str, the time variable

    Returns:

    Raises:

    """
    time_parameter = model.createParameter()
    time_parameter.setId(time_var)
    time_parameter.setName(time_var)
    time_parameter.setConstant(False)

    time_assignment = model.createAssignmentRule()
    time_assignment.setVariable(time_var)
    time_assignment.setMath(sbml.parseL3Formula('time'))


def _read_parameters_block(model: sbml.Model, parameter_list: list):
    """
    Reads and processes the parameters block in the ODE yaml file. In
    particular, it reads the parameters and adds them to the given SBML model.
    The expected format for parameter definition is
    {'parameterId': <parameterId>, 'nominalValue': <nominalValue>}

    Arguments:
        model: the SBML model
        parameter_list: block containing the parameter definitions

    Returns:

    Raises:

    """
    for parameter_def in parameter_list:
        if 'nominalValue' in parameter_def.keys():
            _create_parameter(model,
                              parameter_def['parameterId'],
                              parameter_def['nominalValue'])
        else:
            _create_parameter(model, parameter_def['parameterId'])


def _create_parameter(model: sbml.Model, parameter_id: str, value: str = None):
    """
    Creates a parameter and adds it to the given SBML model.
    Units are set as dimensionless by default.

    Arguments:
        model: the SBML model to which the parameter will be added.
        parameter_id: the parameter ID
        value: the parameter value, if value is None, no parameter is set.

    Returns:

    Raises:

    """
    k = model.createParameter()
    k.setId(parameter_id)
    k.setName(parameter_id)
    k.setConstant(True)

    if value is not None:
        k.setValue(float(value))

    k.setUnits('dimensionless')


def _read_assignments_block(model: sbml.Model, assignment_list: list):
    """
    Reads and processes the assignments block in the ODE yaml file. In
    particular, it reads the assignments and adds them to the given SBML file.
    The expected format of a state definition is:
    {'assignmentId': <assignmentId>, 'formula': <formula>}

    This is used to assign a formula (probably time-dependent) to a variable.

    Arguments:
        model: the SBML model
        assignment_list: a list of dictionaries where each entry is an
                         assignment definition

    Returns:

    Raises:

    """
    for assignment_def in assignment_list:
        _create_assignment(model,
                           assignment_def['assignmentId'],
                           assignment_def['formula'])


def _create_assignment(model: sbml.Model, assignment_id: str, formula: str):
    """
    Creates an  assignment rule, that assigns id to formula.

    Arguments:
        model: SBML model to which the assignment rule will be added.
        assignment_id: str, the id of the assignment rule
        formula: str: contains the equation for the assignment rule
    Returns:

    Raises:

    """
    assignment_parameter = model.createParameter()
    assignment_parameter.setId(assignment_id)
    assignment_parameter.setName(assignment_id)
    assignment_parameter.setConstant(False)
    assignment_parameter.setUnits('dimensionless')

    assignment_rule = model.createAssignmentRule()
    assignment_rule.setVariable(assignment_id)
    assignment_rule.setMath(sbml.parseL3Formula(formula))


def _read_functions_block(model: sbml.Model, functions_list: list):
    """
    Reads and processes the functions block in the ODE yaml file.
    In particular, it reads the functions and adds them to the given SBML file
    as functionDefinitions.
    The expected format of a function definition is:
        {'functionId': <functionId>,
         'arguments': <arguments>,
         'formula' : <formula>}

    Arguments:
        model: a SBML model
        functions_list: a list of dictionaries where each entry is a
                        function definition

    Returns:

    Raises:

    """
    for function_def in functions_list:
        _create_function(model,
                         function_def['functionId'],
                         function_def['arguments'],
                         function_def['formula'])


def _create_function(model: sbml.Model,
                     function_id: str,
                     arguments: str,
                     formula: str):
    """
    Creates a functionDefinition and adds it to the given SBML model.

    Arguments:
        model: SBML model to which the function will be added.
        function_id: the function id/name
        arguments: the arguments of the function (species AND parameters)
        formula: the formula of the function

    Returns:

    Raises:

    """
    f = model.createFunctionDefinition()
    f.setId(function_id)
    math = sbml.parseL3Formula('lambda(' + arguments + ', ' + formula + ')')
    f.setMath(math)


def _read_odes_block(model: sbml.Model, odes_list: list):
    """
    Reads and processes the odes block in the ODE yaml file.
    In particular, it reads the odes and adds a species for the corresponding
    state and the right hand side as rateRules to the given SBML file.

    The expected format of an ode definition is:
    {'stateId': <state_variable>, 'rightHandSide' : <right_hand_side>,
    'initialValue' = <initial_value>}

    Arguments:
        model: a SBML model
        odes_list: block of ODE definitions

    Returns:

    Raises:

    """
    for ode_def in odes_list:
        _create_species(model, ode_def['stateId'], ode_def['initialValue'])
        _create_rate_rule(model, ode_def['stateId'], ode_def['rightHandSide'])


def _create_species(model: sbml.Model, species_id: str, initial_amount: str):
    """
    Creates a species and adds it to the given SBML model.
    Units are set as dimensionless by default.

    Arguments:
        model: the SBML model to which the species will be added.
        species_id: the species ID
        initial_amount: the species initial amount

    Returns:
        s: the SBML species

    Raises:

    """
    s = model.createSpecies()
    s.setId(species_id)

    try:
        s.setInitialAmount(float(initial_amount))
    except ValueError:
        init = model.createInitialAssignment()
        init.setId('init_' + species_id)
        init.setSymbol(species_id)
        init.setMath(sbml.parseL3Formula(initial_amount))

    s.setConstant(False)
    s.setBoundaryCondition(False)
    s.setHasOnlySubstanceUnits(False)
    s.setCompartment('Compartment')

    s.setSubstanceUnits('dimensionless')

    return s


def _create_rate_rule(model: sbml.Model, species: str, formula: str):
    """
    Creates a SBML rateRule for a species and adds it to the given model.
    This is where the ODEs from the text file are encoded.

    Arguments:
        model: SBML model to which the rate rule will be added.
        species: the species name of the ODE
        formula: the right-hand-side of the ODE

    Returns:

    Raises:

    """
    r = model.createRateRule()
    r.setId('d/dt_' + species)
    r.setVariable(species)
    math_ast = sbml.parseL3Formula(formula)
    r.setMath(math_ast)


def _read_observables_block(model: sbml.Model, observable_list: list):
    """
    Reads an processes the observables block in the ODE yaml file.
    Since the observables are not represented in the SBML, it only gives
    a warning to inform the user.

    Arguments:
        model: SBML model (libsbml)
        observable_list: observables block containing all
                         observable definitions.

    Returns:

    Raises:

    """
    warnings.warn('Observables are not represented in the SBML and therefore '
                  'only have an effect the output, when called via yaml2PEtab')


def _read_conditions_block(model: sbml.Model, conditions_list: list):
    """
    Reads an processes the conditions block in the ODE yaml file.
    Since the conditions are not represented in the SBML, it only gives
    a warning to inform the user.

    Arguments:
        model: SBML model (libsbml)
        conditions_list: conditions block containing all
                         conditions definitions.

    Returns:

    Raises:

    """
    warnings.warn('Conditions are not represented in the SBML and therefore '
                  'only have an effect the output, when called via yaml2PEtab')


def main():
    parser = argparse.ArgumentParser(
        description='Takes in an ODE model in .yaml and converts it to SBML.')
    parser.add_argument('yaml_file', type=str, help='Path to input YAML file.')
    parser.add_argument('sbml_file', type=str,
                        help='Path to output SBML file.')

    args = parser.parse_args()

    print(f'Path to yaml file: {args.yaml_file}')
    print(f'Path to sbml file: {args.sbml_file}')

    print('Converting...')

    yaml2sbml(args.yaml_file, args.sbml_file)


if __name__ == '__main__':
    main()



"""Import of yaml2sbml's public API."""

# version
from .version import __version__   # noqa: F401

# API
from yaml2sbml.yaml2sbml import yaml2sbml
from yaml2sbml.yaml2PEtab import yaml2petab, validate_petab_tables
from yaml2sbml.yaml_validation import validate_yaml
from yaml2sbml.YamlModel import YamlModel



import os
import yaml
import jsonschema
import argparse

SCHEMA = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                      "yaml_schema.yaml")


def validate_yaml(yaml_file: str):
    """
    Validates the syntax of the yaml file.

    Arguments:
        yaml_file: path to yaml file to be validated

    Returns:
        jsonschema.validate

    Raises:

    """
    # read in yaml_file
    with open(yaml_file, 'r') as f_in:
        yaml_contents = f_in.read()
        yaml_in = yaml.full_load(yaml_contents)

    _validate_yaml_from_dict(yaml_in)
    print('YAML file is valid ✅')


def _validate_yaml_from_dict(yaml_dict: dict):
    """
    Validates the syntax of the yaml file, using a dict as input.

    Arguments:
        yaml_dict: yaml model as dict.

    Returns:
        jsonschema.validate

    Raises:

    """
    # read in SCHEMA
    with open(SCHEMA, 'r') as f_in:
        yaml_contents = f_in.read()
        schema = yaml.full_load(yaml_contents)

    jsonschema.validate(instance=yaml_dict, schema=schema)


def main():
    """
    Function called by the CLI.
    """
    parser = argparse.ArgumentParser(
        description='Validates a yaml model '
                    'so that it can be used by yaml2bsml.')

    parser.add_argument('yaml_file', type=str,
                        help='Directory of yaml file, that '
                             'should be validated.')

    args = parser.parse_args()

    print(f'Path to yaml file: {args.yaml_file}')
    print('Validating...')

    validate_yaml(args.yaml_file)


if __name__ == '__main__':
    main()



import os
import unittest

from yaml2sbml.yaml2sbml import _parse_yaml


class TestYaml2SBML(unittest.TestCase):
    """
    TestCase class for testing ODE import from a generic yaml file
    and conversion to SBML.
    """

    def setUp(self):
        this_dir, _ = os.path.split(__file__)
        self.test_folder = os.path.join(this_dir, 'test_yaml2sbml')

    def test_yaml_import(self):
        """
        Test yaml import/SBML generation...
        """
        ode_file = os.path.join(self.test_folder, 'ode_input1.yaml')
        expected_result_file = os.path.join(self.test_folder,
                                            'true_sbml_output.xml')

        sbml_contents = _parse_yaml(ode_file)

        with open(expected_result_file, 'r') as f_in:
            expected_sbml_contents = f_in.read()

        sbml_test_dir = os.path.join(self.test_folder, 'sbml_test.xml')
        with open(sbml_test_dir, 'w') as f_out:
            f_out.write(sbml_contents)

        self.assertEqual(expected_sbml_contents, sbml_contents)

        os.remove(sbml_test_dir)

    def test_yaml_import_observables(self):
        """
        Test yaml import/export for a model containing observables
        (that are not translated).
        """
        ode_file = os.path.join(self.test_folder, 'ode_input2.yaml')

        expected_result_file = \
            os.path.join(self.test_folder, 'true_sbml_output.xml')

        test_sbml_dir = os.path.join(self.test_folder, 'sbml_test.xml')

        sbml_contents = _parse_yaml(ode_file)

        with open(expected_result_file, 'r') as f_in:
            expected_sbml_contents = f_in.read()

        with open(test_sbml_dir, 'w') as f_out:
            f_out.write(sbml_contents)

        self.assertEqual(expected_sbml_contents, sbml_contents)

        os.remove(test_sbml_dir)


if __name__ == '__main__':
    suite = unittest.TestSuite()
    suite.addTest(TestYaml2SBML())
    unittest.main()



import yaml
import os.path
from typing import Union
import copy

from .yaml2sbml import _parse_yaml_dict
from .yaml2PEtab import _yaml2petab
from .yaml_validation import _validate_yaml_from_dict


class YamlModel:
    """
    Functionality to set up, edit, load and write yaml models.
    """

    def __init__(self):
        """
        Set up yaml model.
        """
        self._yaml_model = {'time': {},
                            'odes': [],
                            'parameters': [],
                            'assignments': [],
                            'functions': [],
                            'observables': [],
                            'conditions': []}

    @staticmethod
    def load_from_yaml(yaml_file):
        """
        Creates a model instance from a yaml file.

        Arguments:
        yaml_dir:
            directory to the yaml file, that should be imported

        Returns:
        cls:
            new model
        """

        new_model = YamlModel()

        # read in yaml_file
        with open(yaml_file, 'r') as f_in:
            yaml_contents = f_in.read()
            new_model._yaml_model.update(yaml.full_load(yaml_contents))

        # check, if the model is valid
        new_model.validate_model()

        return new_model

    def write_to_yaml(self,
                      yaml_dir: str,
                      overwrite: bool = False):
        """
        Write model to yaml file given as directory yaml_dir.

        Arguments:
        yaml_dir:
            path/file, where the yaml should be written
        overwrite:
            Indicates, whether an existing yaml should be overwritten

        Returns:

        Raises:
            ValueError
            FileExistsError
        """
        if not (yaml_dir.endswith('.yaml') or yaml_dir.endswith('.yml')):
            raise ValueError('yaml_dir should contain path to the yaml '
                             'and hence end with .yaml or .yml')

        if (not overwrite) and os.path.exists(yaml_dir):
            raise FileExistsError(f'Can not write yaml model. File {yaml_dir}'
                                  f' already exists. Consider to set '
                                  f'overwrite=True.')

        reduced_model_dict = self._get_reduced_model_dict()

        # translate to string
        yaml_as_string = yaml.dump(reduced_model_dict,
                                   sort_keys=False,
                                   indent=6)

        # post-process: add empty line around blocks
        for key in self._yaml_model.keys():
            yaml_as_string = yaml_as_string.replace(f'{key}:',
                                                    f'\n{key}:')
        yaml_as_string = yaml_as_string.replace('-     ', '\n    - ')

        with open(yaml_dir, 'w') as file:
            file.write(yaml_as_string)

    def write_to_sbml(self,
                      sbml_dir: str,
                      overwrite: bool = False):
        """
        Writes the model as an SBML file to the directory given in sbml_dir.

        Arguments:
        sbml_dir:
            path/file, where the sbml should be written
        overwrite:
            Indicates, whether an existing yaml should be overwritten

        Returns:

        Raises:
            ValueError
            FileExistsError
        """
        # Check file ending.
        if not (sbml_dir.endswith('.xml') or sbml_dir.endswith('.sbml')):
            raise ValueError('sbml_dir should contain path to the sbml '
                             'and hence end with .xml or .sbml')

        if (not overwrite) and os.path.exists(sbml_dir):
            raise FileExistsError(f'Can not write SBML model. File {sbml_dir}'
                                  f' already exists. Consider to set '
                                  f'overwrite=True.')

        # generate SBML as string
        reduced_model_dict = self._get_reduced_model_dict()
        sbml_as_string = _parse_yaml_dict(reduced_model_dict)

        with open(sbml_dir, 'w') as f_out:
            f_out.write(sbml_as_string)

    def write_to_petab(self,
                       output_dir: str,
                       model_name: str,
                       petab_yaml_name: str = None,
                       measurement_table_name: str = None):
        """
        Writes the YamlModel as a petab problem. Equivalent to calling
        yaml2petab on the file produced by the yaml output.

        If a petab_yaml_name is given, a .yaml file is created, that organizes
        the petab problem. If additionally a measurement_table_file_name is
        specified, this file name is written into the created .yaml file.

        Arguments:
            output_dir: path the output file(s) are be written out
            model_name: name of SBML model
            petab_yaml_name: name of yaml organizing the PEtab problem.
            measurement_table_name: Name of measurement table
        """
        reduced_model_dict = self._get_reduced_model_dict()

        _yaml2petab(reduced_model_dict,
                    output_dir,
                    model_name,
                    petab_yaml_name,
                    measurement_table_name)

    def validate_model(self):
        """
        Validates the yaml model.

        Arguments:

        Returns:

        Raises:
            ValidationError
        """
        _validate_yaml_from_dict(self._yaml_model)

    def _get_reduced_model_dict(self) -> dict:
        """
        Returns a reduced model dict, where keys without an entry are deleted.
        (Returns a copy of the model dict!!)

        Arguments:

        Raises:

        Returns:
            reduced_model_dict
        """
        reduced_model_dict = {}

        for (key, val) in self._yaml_model.items():
            if val:
                reduced_model_dict[key] = copy.deepcopy(val)

        return reduced_model_dict

    # functionalities regarding the time
    def is_set_time(self):
        return 'variable' in self._yaml_model['time'].keys()

    def set_time(self,
                 time_variable: str):
        self._yaml_model['time'] = {'variable': time_variable}

    def delete_time(self):
        self._yaml_model['time'] = {}

    def get_time(self):
        if self.is_set_time():
            return self._yaml_model['time']['variable']
        else:
            return None

    # functions adding a value
    def add_parameter(self,
                      parameter_id: str,
                      overwrite: bool = False,
                      nominal_value: float = None,
                      parameter_name: str = None,
                      parameter_scale: str = None,
                      lower_bound: float = None,
                      upper_bound: float = None,
                      estimate: float = None):
        """
        Adds a parameter. Overwrites an existing parameter with the same id,
        if overwrite=True.
        """
        # if parameter exists: delete if overwrite
        if parameter_id in self.get_parameter_ids():
            if overwrite:
                self.delete_parameter(parameter_id)
            else:
                raise ValueError('Could not add parameter with id'
                                 f' {parameter_id}: Parameter with the same'
                                 ' id already exists.')

        entry_dict = {'parameterId': parameter_id,
                      'nominalValue': nominal_value,
                      'parameterName': parameter_name,
                      'parameterScale': parameter_scale,
                      'lowerBound': lower_bound,
                      'upperBound': upper_bound,
                      'estimate': estimate}

        self._add_entry(entry_dict, 'parameters')

    def add_ode(self,
                state_id: str,
                right_hand_side: Union[float, str],
                initial_value: Union[float, str],
                overwrite: bool = False):
        """
        Adds a state/ODE. Overwrites an existing state/ODE with the same id,
        if overwrite=True.
        """
        # if state exists: delete if overwrite
        if state_id in self.get_ode_ids():
            if overwrite:
                self.delete_ode(state_id)
            else:
                raise ValueError(f'Could not add state/ODE with id {state_id}:'
                                 f' State with the same id already exists.')

        entry_dict = {'stateId': state_id,
                      'rightHandSide': right_hand_side,
                      'initialValue': initial_value}

        self._add_entry(entry_dict, 'odes')

    def add_assignment(self,
                       assignment_id: str,
                       formula: str,
                       overwrite: bool = False):
        """
        Adds an assignment. Overwrites an existing assignment with the same id,
        if overwrite=True.
        """
        # if assignment exists: delete if overwrite
        if assignment_id in self.get_assignment_ids():
            if overwrite:
                self.delete_assignment(assignment_id)
            else:
                raise ValueError('Could not add assignment with id '
                                 f'{assignment_id}: Assignment with the same '
                                 'id already exists.')

        entry_dict = {'assignmentId': assignment_id,
                      'formula': formula}

        self._add_entry(entry_dict, 'assignments')

    def add_function(self,
                     function_id: str,
                     arguments: str,
                     formula: str,
                     overwrite: bool = False):
        """
        Adds a function. Overwrites an existing function with the same id,
        if overwrite=True.
        """
        # if function exists: delete if overwrite
        if function_id in self.get_function_ids():
            if overwrite:
                self.delete_function(function_id)
            else:
                raise ValueError('Could not add function with id '
                                 f' {function_id}: Function with the same '
                                 'id already exists.')

        entry_dict = {'functionId': function_id,
                      'arguments': arguments,
                      'formula': formula}

        self._add_entry(entry_dict, 'functions')

    def add_observable(self,
                       observable_id: str,
                       observable_formula: str,
                       noise_formula: str,
                       overwrite: bool = False,
                       observable_name: str = None,
                       observable_transformation: str = None,
                       noise_distribution: str = None):
        """
        Adds an observable. Overwrites an existing observable with the same id,
        if overwrite=True.
        """
        # if observable exists: delete if overwrite
        if observable_id in self.get_observable_ids():
            if overwrite:
                self.delete_observable(observable_id)
            else:
                raise ValueError('Could not add observable with id '
                                 f'{observable_id}: Observable with the same '
                                 'id already exists.')

        entry_dict = {'observableId': observable_id,
                      'observableName': observable_name,
                      'observableFormula': observable_formula,
                      'observableTransformation': observable_transformation,
                      'noiseFormula': noise_formula,
                      'noiseDistribution': noise_distribution}

        self._add_entry(entry_dict, 'observables')

    def add_condition(self,
                      condition_id: str,
                      condition_dict: dict,
                      overwrite: bool = False,
                      condition_name: str = None):

        """
        Adds a condition. Overwrites an existing condition with the same id,
        if overwrite=True.

        Arguments:
        condition_id:
            str, condition id
        condition_dict:
            dict, of the form {<parameter or state id>: <value>}.
            Corresponds to entries in the PEtab condition table.
            See details there.
        overwrite:
            bool, indicates if an existing condition should be overwritten
        condition_name:
            Condition name. Optional.

        Returns:
        """
        # if condition exists: delete if overwrite
        if condition_id in self.get_condition_ids():
            if overwrite:
                self.delete_condition(condition_id)
            else:
                raise ValueError('Could not add condition with id '
                                 f' {condition_id}: Condition with the same '
                                 f'id already exists.')

        entry_dict = {'conditionId': condition_id,
                      'conditionName': condition_name,
                      **condition_dict}

        self._add_entry(entry_dict, 'conditions')

    def _add_entry(self,
                   entry_dict: dict,
                   block_key: str):
        """
        Adds the entry in 'entry_dict' to the block indexed by 'block_key'.
        If 'overwrite=True', an existing value for that key is overwritten.

        Arguments:
        entry_dict:
            dict, that stores the new entry
        block_key:
            name, where the ids should be searched (e.g. 'parameters')

        Returns:
        """
        # filter out None values and append
        filtered_dict = _filter_none_values(entry_dict)
        self._yaml_model[block_key].append(filtered_dict)

    # functionalities to get ids
    def get_parameter_ids(self):
        """Returns a list with all parameter ids."""
        return self._get_ids('parameters', 'parameterId')

    def get_ode_ids(self):
        """Returns a list with all state ids."""
        return self._get_ids('odes', 'stateId')

    def get_assignment_ids(self):
        """Returns a list with all assignment ids."""
        return self._get_ids('assignments', 'assignmentId')

    def get_function_ids(self):
        """Returns a list with all function ids."""
        return self._get_ids('functions', 'functionId')

    def get_observable_ids(self):
        """Returns a list with all observable ids."""
        return self._get_ids('observables', 'observableId')

    def get_condition_ids(self):
        """Returns a list with all conditions ids."""
        return self._get_ids('conditions', 'conditionId')

    def _get_ids(self,
                 block_key: str,
                 index_key: str):
        """
        Returns all ids in the block named `block_key` in self._yaml_model.
        The ids have the key index_key.

        Arguments:
        block_key:
            name, where the ids should be searched (e.g. 'parameters')
        index_key:
            key of the identifier, in that block (e.g. 'parameterId')

        Returns:
        res:
            list of ids
        """
        return [element[index_key] for element in self._yaml_model[block_key]]

    # functionalities to get entry by Id:
    def get_parameter_by_id(self,
                            parameter_id: str):
        """returns dict for corresponding parameter."""
        if parameter_id not in self.get_parameter_ids():
            raise IndexError(f'Could not find parameter {parameter_id}.')

        return self._get_entry_by_id('parameters',
                                     'parameterId',
                                     parameter_id)

    def get_ode_by_id(self,
                      state_id: str):
        """returns dict for corresponding ODE/state."""
        if state_id not in self.get_ode_ids():
            raise IndexError(f'Could not find state/ODE {state_id}.')

        return self._get_entry_by_id('odes',
                                     'stateId',
                                     state_id)

    def get_assignment_by_id(self,
                             assignment_id: str):
        """returns dict for corresponding assignment."""
        if assignment_id not in self.get_assignment_ids():
            raise IndexError(f'Could not find assignment {assignment_id}.')

        return self._get_entry_by_id('assignments',
                                     'assignmentId',
                                     assignment_id)

    def get_function_by_id(self,
                           function_id: str):
        """returns dict for corresponding function."""
        if function_id not in self.get_function_ids():
            raise IndexError(f'Could not find function {function_id}.')

        return self._get_entry_by_id('functions',
                                     'functionId',
                                     function_id)

    def get_observable_by_id(self,
                             observable_id: str):
        """returns dict for corresponding observable."""
        if observable_id not in self.get_observable_ids():
            raise IndexError(f'Could not find observable {observable_id}.')

        return self._get_entry_by_id('observables',
                                     'observableId',
                                     observable_id)

    def get_condition_by_id(self,
                            condition_id: str):
        """returns dict for corresponding condition."""
        if condition_id not in self.get_condition_ids():
            raise IndexError(f'Could not find condition {condition_id}.')

        return self._get_entry_by_id('conditions',
                                     'conditionId',
                                     condition_id)

    def _get_entry_by_id(self,
                         block_key: str,
                         index_key: str,
                         entry_id: str):
        """
        Returns the entry with id 'entry_id' in the block named
        `block_key` in self._yaml_model. 'index_key' gives the name of the
        id-key.

        Arguments:
        block_key:
            Key of the block (e.g. 'parameters')
        index_key:
            key of the identifier in that block (e.g. 'parameterId')
        entry_id:
            Id of element, that should be returned

        Returns:
            Entry_dict. None, if entry is not found.
        """
        # check if block exists.
        for entry in self._yaml_model[block_key]:
            if entry[index_key] == entry_id:
                return entry

        return None

    # functionalities to delete entry by Id:
    def delete_parameter(self,
                         parameter_id: str):
        """
        Deletes a parameter. Raises a ValueError, if parameter does not exist.
        """
        if not self._delete_entry('parameters',
                                  'parameterId',
                                  parameter_id):

            raise ValueError(f'Could not delete parameter {parameter_id}. '
                             f'Invalid parameterId.')

    def delete_ode(self,
                   state_id: str):
        """
        Deletes a state + ODE. Raises a ValueError, if state does not exist.
        """
        if not self._delete_entry('odes',
                                  'stateId',
                                  state_id):

            raise ValueError(f'Could not delete ODE for state {state_id}. '
                             f'Invalid stateId.')

    def delete_assignment(self,
                          assignment_id: str):
        """
        Deletes an assignment. Raises a ValueError,
        if assignment does not exist.
        """
        if not self._delete_entry('assignments',
                                  'assignmentId',
                                  assignment_id):

            raise ValueError(f'Could not delete assignment {assignment_id}. '
                             f'Invalid assignmentId.')

    def delete_function(self,
                        function_id: str):
        """
        Deletes a function. Raises a ValueError, if function does not exist.
        """
        if not self._delete_entry('functions',
                                  'functionId',
                                  function_id):

            raise ValueError(f'Could not delete function {function_id}. '
                             f'Invalid functionId.')

    def delete_observable(self,
                          observable_id: str):
        """
        Deletes an observable. Raises a ValueError,
        if observable does not exist.
        """
        if not self._delete_entry('observables',
                                  'observableId',
                                  observable_id):

            raise ValueError(f'Could not delete observable {observable_id}. '
                             f'Invalid observableId.')

    def delete_condition(self,
                         condition_id: str):
        """
        Deletes a condition. Raises a ValueError, if condition does not exist.
        """
        if not self._delete_entry('conditions',
                                  'conditionId',
                                  condition_id):

            raise ValueError(f'Could not delete condition {condition_id}. '
                             f'Invalid conditionId.')

    def _delete_entry(self,
                      block_key: str,
                      index_key: str,
                      deleted_object_id: str):
        """
        Deletes the entry with id 'deleted_object_id' in the block named
        `block_key` in self._yaml_model. 'index_key' gives the name of the
        id-key.

        Arguments:
        block_key:
            Key of the block (e.g. 'parameters')
        index_key:
            key of the identifier in that block (e.g. 'parameterId')
        deleted_object_id:
            Id of element, that should be deleted

        Returns:
            Bool, that indicates, whether deletion was successful.
        """

        # Check if block exists
        for i, entry in enumerate(self._yaml_model[block_key]):

            # search for entry and delete
            if entry[index_key] == deleted_object_id:
                self._yaml_model[block_key].pop(i)
                return True

        return False


def _filter_none_values(d: dict):
    """
    Filters out the key-value pairs with None  as value.

    Arguments:
    d
        dictionary

    Returns:
        filtered dictionary.
    """
    return {key: value for (key, value) in d.items() if value is not None}



import os
import unittest

import yaml2sbml.yaml2PEtab as yaml2PEtab


class TestYaml2PEtab(unittest.TestCase):
    """
    TestCase class for testing ODE import from a generic yaml file
    and conversion to PEtab.
    """

    def setUp(self):
        this_dir, _ = os.path.split(__file__)
        self.test_folder = os.path.join(this_dir, 'test_yaml2sbml')

    def test_petab_export(self):
        """
        Test PEtab export
        """
        ode_file = os.path.join(self.test_folder, 'ode_input2.yaml')

        yaml2PEtab.yaml2petab(ode_file,
                              self.test_folder,
                              'sbml_test.xml')

        yaml2PEtab.validate_petab_tables(
            os.path.join(self.test_folder, 'sbml_test.xml'), self.test_folder)

        for f in ['observable_table.tsv',
                  'parameter_table.tsv',
                  'condition_table.tsv',
                  'sbml_test.xml']:
            os.remove(os.path.join(self.test_folder, f))


if __name__ == '__main__':
    suite = unittest.TestSuite()
    suite.addTest(TestYaml2PEtab())
    unittest.main()



import setuptools
import os

ENTRY_POINTS = {
    'console_scripts': [
        'yaml2sbml = yaml2sbml.yaml2sbml:main',
        'yaml2petab = yaml2sbml.yaml2PEtab:main',
        'yaml2sbml_validate = yaml2sbml.yaml_validation:main'
    ]
}

with open(os.path.join(os.path.dirname(__file__),
          "yaml2sbml", "version.py")) as f:
    version = f.read().split('\n')[0].split("'")[-2]

setuptools.setup(
    name="yaml2sbml",
    version=version,
    author="Jakob Vanhoefer, Marta R. A. Matos",
    author_email="marta.ra.matos@gmail.com",
    description="A small package to convert ODEs specified in "
                "a yaml file to SBML/PEtab.",
    url="https://github.com/martamatos/yaml2sbml",
    packages=setuptools.find_packages(),
    install_requires=["python-libsbml>=5.18.0",
                      "PyYAML>=5.1",
                      "pandas>=1.0.1",
                      "petab>=0.1.4"],
    tests_require=["amici>=0.11.10",
                   "pypesto>=0.2.2"
                   "numpy>=1.19.4",
                   "matplotlib>=3.1.0",
                   "flake8>=3.7.2",
                   "nbmake>=0.1.0",
                   "scipy>=1.6.0"],
    extras_require={'examples': ["amici>=0.11.10",
                                 "numpy>=1.19.4",
                                 "matplotlib>=3.1.0",
                                 "scipy>=1.6.0"]},
    python_requires='>=3.6',
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
    ],
    entry_points=ENTRY_POINTS,
    include_package_data=True
)
